/* BBHotSpots - a BB4W plugin to add `hotspots'.
   Copyright (C) 2005 kana <nicht AT s8 DOT xrea DOT con>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
	/* for WM_XBUTTONUP */
#define _WIN32_WINNT 0x0500
#define WINVER 0x0500

#include <windows.h>
#include <shellapi.h>

#include <BBApi-C.h>
#include <BBCmd.h>

#ifdef __GNUC__
#define PRINTF_LIKE(string_index, first_to_check) \
        __attribute__((format(printf, string_index, first_to_check)))
#else
#define PRINTF_LIKE(string_index, first_to_check)
#endif


#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))




/* Constants */

static const char BBHS_NAME[] = "BBHotSpots";
static const char BBHS_VERSION[] = "0.0.3";
static const char BBHS_AUTHOR[] = "kana";
static const char BBHS_RELEASE[] = "2005-01-23/2006-01-30";
static const char BBHS_LINK[] = "http://nicht.s8.xrea.com/2005/01/BBHotSpots";
static const char BBHS_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static const char BBHS_CLASS_NAME[] = "BBHotSpots";
static const char BBHS_RC_NAME[] = "BBHotSpots.rc";
static const UINT BBHS_MESSAGES[] = {BB_BROADCAST, BB_RECONFIGURE, 0};

#define WH_MOUSE_CALLBACK (WM_APP + 0)

#define IDT_CHECK 1




/* Global variables */

static HWND G_PluginWnd = NULL;
static HMODULE G_PluginModule = NULL;
static RECT G_ScreenRect;


#define HOTSPOT_NAME_MAX 80
typedef struct _HotSpot HotSpot;
struct _HotSpot{
	char* name;
	RECT spot;
	BBCmdCommand* enter_script;
	BBCmdCommand* leave_script;
	BBCmdCommand* click_script;
	BOOL active;
	HotSpot* next;
};

static HotSpot* G_HotSpotList = NULL;

static enum {
	TIMER,
	HOOK
} G_Engine = TIMER;

static int G_CheckingInterval = 1000;

static HMODULE G_HookDLLModule = NULL;


static HotSpot* G_LastActiveHotSpot = NULL;
static POINT G_LastPosition = {-1, -1};








/* Misc. functions */

static void
VSNPRINTF(char* buf, int size, const char* format, va_list va)
{
	if (buf == NULL || size <= 0)
		return;

	vsnprintf(buf, size, format, va);

	buf[size - 1] = '\0';
}


static char*
STRNCPY(char* dest, const char* src, size_t n)
{
	strncpy(dest, src, n);
	dest[n - 1] = '\0';

	return dest;
}


static char*
strndup(const char* s, int n)
{
	char* p;

	p = malloc(sizeof(char) * (n+1));
	if (p != NULL)
		STRNCPY(p, s, n+1);

	return p;
}


static void
tokenize( const char* string,
          const char* token[],
          int token_length[],
          int* token_num,
          int max,
          const char** rest )
{
	const char* WS = " \t";  /* WhiteSpace */
	const char* head;
	const char* tail;
	int n;

	head = string + strspn(string, WS);
	n = 0;
	while ((*head != '\0') && (n < max)) {
		head += strspn(head, WS);

		if (*head == '\'') {
			head++;
			tail = head + strcspn(head, "'");
		} else if (*head == '"') {
			head++;
			tail = head;
			while ((*tail != '\0') && (*tail != '"')) {
				if ((tail[0] == '\\') && (tail[1] != '\0'))
					tail++;
				tail++;
			}
		} else {
			tail = head + strcspn(head, WS);
		}

		token[n] = head;
		token_length[n] = (tail - head) / sizeof(char);

		head = tail + ((*tail == '\'') || (*tail == '"') ? 1 : 0);
		n++;
	}

	*token_num = n;
	*rest = head + strspn(head, WS);
}




PRINTF_LIKE(1, 2)
static void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	VSNPRINTF(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( G_PluginWnd, buf, BBHS_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TOPMOST);
}




static BOOL
is_2k_or_later(void)
{
	OSVERSIONINFO os;

	os.dwOSVersionInfoSize = sizeof(os);
	GetVersionEx(&os);

	return 5 <= os.dwMajorVersion;
}


static BOOL
is_98_or_2k_or_later(void)
{
	OSVERSIONINFO os;

	os.dwOSVersionInfoSize = sizeof(os);
	GetVersionEx(&os);

	return (5 <= os.dwMajorVersion)
	       || ((4 <= os.dwMajorVersion) && (10 <= os.dwMinorVersion));
}




static void
update_screen_rect(void)
{
	if (is_98_or_2k_or_later()) {
		G_ScreenRect.left = GetSystemMetrics(SM_XVIRTUALSCREEN);
		G_ScreenRect.top = GetSystemMetrics(SM_YVIRTUALSCREEN);
		G_ScreenRect.right = GetSystemMetrics(SM_CXVIRTUALSCREEN);
		G_ScreenRect.bottom = GetSystemMetrics(SM_CYVIRTUALSCREEN);
	} else {
		G_ScreenRect.left = 0;
		G_ScreenRect.top = 0;
		G_ScreenRect.right = GetSystemMetrics(SM_CXSCREEN);
		G_ScreenRect.bottom = GetSystemMetrics(SM_CYSCREEN);
	}
}


static LONG
convert_width(const char* width, int length)
{
	const char* p;

	if (((p = strchr(width, '%')) != NULL) && (p < width + length))
		return G_ScreenRect.right * atoi(width) / 100;
	else
		return atoi(width);
}

static LONG
convert_height(const char* height, int length)
{
	const char* p;

	if (((p = strchr(height, '%')) != NULL) && (p < height + length))
		return G_ScreenRect.bottom * atoi(height) / 100;
	else
		return atoi(height);
}

static LONG
convert_x(const char* x, int length, LONG width)
{
	const char* p;

	if (((p = strchr(x, '%')) != NULL) && (p < x + length))
		return G_ScreenRect.left + G_ScreenRect.right * atoi(x) / 100;
	else if (!strnicmp(x, "left", length))
		return G_ScreenRect.left;
	else if (!strnicmp(x, "center", length))
		return G_ScreenRect.left + G_ScreenRect.right / 2 - width / 2;
	else if (!strnicmp(x, "right", length))
		return G_ScreenRect.left + G_ScreenRect.right - width;
	else {
		int v;

		v = atoi(x);
		if (v < 0)
			return G_ScreenRect.left + G_ScreenRect.right + v;
		else
			return G_ScreenRect.left + v;
	}
}

static LONG
convert_y(const char* y, int length, LONG height)
{
	const char* p;

	if (((p = strchr(y, '%')) != NULL) && (p < y + length))
		return G_ScreenRect.top + G_ScreenRect.bottom * atoi(y) / 100;
	else if (!strnicmp(y, "top", length))
		return G_ScreenRect.top;
	else if (!strnicmp(y, "center", length))
		return G_ScreenRect.top + G_ScreenRect.bottom / 2 - height / 2;
	else if (!strnicmp(y, "bottom", length))
		return G_ScreenRect.top + G_ScreenRect.bottom - height;
	else {
		int v;

		v = atoi(y);
		if (v < 0)
			return G_ScreenRect.left + G_ScreenRect.right + v;
		else
			return G_ScreenRect.left + v;
	}
}


static BOOL
convert_bool(const char* s, int length)
{
	return !strnicmp(s, "true", length);
}




static BOOL
get_module_dir(char* buf, int size)
{
	char* bs;

	if (!GetModuleFileName(G_PluginModule, buf, size))
		return FALSE;
	buf[size - 1] = '\0';

	bs = strrchr(buf, '\\');
	if (bs != NULL)
		bs[1] = '\0';

	return TRUE;
}


static BOOL
get_file_path(char* buf, int size, const char* filename)
{
	char dir[MAX_PATH];

	if (!get_module_dir(dir, NUMBER_OF(dir)))
		return FALSE;

	if (size < strlen(dir) + strlen(filename) + 1)
		return FALSE;

	strcpy(buf, dir);
	strcat(buf, filename);

	return TRUE;
}




static BBCmdCommand*
compile(const char* script, int* line, int* column, const char** message)
{
	BBCmdCommand* cmd;
	BBCmdCompiler* compiler;

	compiler = BBCmdCompiler_New();
	if (compiler == NULL) {
		*line = 0;
		*column = 0;
		*message = "Out of memory (BBCmdCompiler_New failed)";
		return NULL;
	}

	cmd = BBCmdCompiler_CompileFromString( compiler, script,
	                                       line, column, message );

	BBCmdCompiler_Del(compiler);
	return cmd;
}




static BOOL
read_settings(void)
{
	char rc[MAX_PATH];
	FILE* file;
	char line[256];
	char* nl;

	if (!get_file_path(rc, NUMBER_OF(rc), BBHS_RC_NAME)) {
		notice("Failed to load %s (too long path)", BBHS_RC_NAME);
		return FALSE;
	}

	file = fopen(rc, "rt");
	if (file == NULL) 
		return TRUE;

	while (fgets(line, NUMBER_OF(line), file) != NULL) {
		if (line[0] != '@')
			continue;

		nl = strchr(line, '\n');
		if (nl != NULL)
			*nl = '\0';

		SendMessage(GetBBWnd(), BB_BROADCAST, 0, (LPARAM)line);
	}

	fclose(file);
	return TRUE;
}




static HotSpot*
HotSpot_New(const char* name, int name_length, const RECT* spot, BOOL active)
{
	HotSpot* self;

	self = malloc(sizeof(*self));
	if (self == NULL)
		return self;

	self->name = strndup(name, name_length);
	if (self->name == NULL) {
		free(self);
		return NULL;
	}
	self->spot = *spot;
	self->enter_script = NULL;
	self->leave_script = NULL;
	self->click_script = NULL;
	self->active = active;
	self->next = NULL;

	return self;
}


static void
HotSpot_Del(HotSpot* self)
{
	free(self->name);
	BBCmdCommand_Del(self->enter_script);
	BBCmdCommand_Del(self->leave_script);
	BBCmdCommand_Del(self->click_script);
	free(self);
}




static HotSpot*
HotSpotList_New(void)
{
	return NULL;
}


static void
HotSpotList_Del(HotSpot* self)
{
	HotSpot* hotspot;
	HotSpot* next;

	hotspot = self;
	while (hotspot != NULL) {
		next = hotspot->next;
		HotSpot_Del(hotspot);
		hotspot = next;
	}
	return;
}


static int
HotSpotList_Add( HotSpot** self, const char* name, int name_length,
                 const RECT* spot, BOOL active )
{
	HotSpot* hotspot;

	hotspot = HotSpot_New(name, name_length, spot, active);
	if (hotspot == NULL)
		return ERROR_OUTOFMEMORY;

	if (*self == NULL)
		*self = hotspot;
	else {
		HotSpot* t;

		t = *self;
		while (TRUE) {
			if ( (strlen(t->name) == name_length)
			     && !strncmp(t->name, name, name_length) )
			{
				HotSpot_Del(hotspot);
				return ERROR_ALREADY_EXISTS;
			}

			if (t->next == NULL)
				break;
			t = t->next;
		}

		t->next = hotspot;
	}
	return NO_ERROR;
}


static BOOL
HotSpotList_DelByName(HotSpot** self, const char* name, int name_length)
{
	HotSpot* hotspot;
	HotSpot* prev;

	hotspot = *self;
	prev = NULL;
	while (hotspot != NULL) {
		if ( (strlen(hotspot->name) == name_length)
		     && !strncmp(hotspot->name, name, name_length) )
		{
			if (prev == NULL)
				*self = hotspot->next;
			else
				prev->next = hotspot->next;
			HotSpot_Del(hotspot);

			return TRUE;
		}

		prev = hotspot;
		hotspot = hotspot->next;
	}

	return FALSE;
}


static HotSpot*
HotSpotList_SearchByName(HotSpot* self, const char* name, int name_length)
{
	HotSpot* hotspot;

	hotspot = self;
	while (hotspot != NULL) {
		if ( (strlen(hotspot->name) == name_length)
		     && !strncmp(hotspot->name, name, name_length) )
		{
			return hotspot;
		}

		hotspot = hotspot->next;
	}

	return NULL;
}


typedef BOOL (*HotSpotList_ForEachMatch)(HotSpot*, void*);
typedef BOOL (*HotSpotList_ForEachProc)(HotSpot*, void*);

static int
HotSpotList_ForEach( HotSpot* self,
                     HotSpotList_ForEachMatch match, void* mparam,
                     HotSpotList_ForEachProc proc, void* pparam )
{
	HotSpot* hotspot;
	int n;

	hotspot = self;
	n = 0;
	while (hotspot != NULL) {
		if ((*match)(hotspot, mparam)) {
			n++;
			if (!(*proc)(hotspot, pparam))
				return n;
		}

		hotspot = hotspot->next;
	}

	return n;
}




typedef struct {
	const char* string;
	int length;
} Token;


static BOOL
HotSpot_MatchByName(HotSpot* self, void* param)
{
	Token* token = param;

	return (strlen(self->name) == token->length)
	       && !strncmp(self->name, token->string, token->length);
}


static BOOL
HotSpot_SetSpot(HotSpot* self, void* param)
{
	self->spot = *(RECT*)param;
	return FALSE;
}


static BOOL
HotSpot_Activate(HotSpot* self, void* param)
{
	self->active = TRUE;
	return (BOOL)param;
}


static BOOL
HotSpot_Inactivate(HotSpot* self, void* param)
{
	self->active = FALSE;
	return (BOOL)param;
}


static BOOL
HotSpot_Toggle(HotSpot* self, void* param)
{
	self->active = !(self->active);
	return (BOOL)param;
}


static BOOL
HotSpot_Info(HotSpot* self, void* param)
{
	notice( "Name:   %s\n"
	        "State:  %s\n"
	        "X:      %ld\n"
	        "Y:      %ld\n"
	        "Width:  %ld\n"
	        "Height: %ld\n"
	        "Enter:  %c\n"
	        "Leave:  %c\n"
	        "Click:  %c",
	        self->name,
	        (self->active ? "active" : "inactive"),
	        self->spot.left,
	        self->spot.top,
	        self->spot.right,
	        self->spot.bottom,
	        (self->enter_script == NULL ? 'x' : 'o'),
	        (self->leave_script == NULL ? 'x' : 'o'),
	        (self->click_script == NULL ? 'x' : 'o') );
	return (BOOL)param;
}








/* Main functions */

static BOOL
engine_begin(void)
{
	if (G_Engine == TIMER) {
		if (!SetTimer(G_PluginWnd,IDT_CHECK,G_CheckingInterval,NULL)) {
			notice("SetTimer failed");
			return FALSE;
		}
		return TRUE;
	} else {
		char path[MAX_PATH];
		FARPROC end;
		FARPROC begin;

		if (!get_file_path(path,NUMBER_OF(path),"BBHSMouseHook.dll")) {
		  notice("Failed to load BBHSMouseHook.dll: Too long path");
		  goto E_PATH;
		}

		G_HookDLLModule = LoadLibrary(path);
		if (G_HookDLLModule == NULL) {
			notice("Failed to load BBHSMouseHook.dll");
			goto E_LOAD;
		}

		begin = GetProcAddress(G_HookDLLModule, "BBHSMouseHook_Begin");
		end = GetProcAddress(G_HookDLLModule, "BBHSMouseHook_End");
		if ((begin == NULL) || (end == NULL)) {
			notice("Invalid BBHSMouseHook.dll");
			goto E_PROC;
		}

		if ( !(*((BOOL (*)(HWND, UINT))begin))
		        (G_PluginWnd, WH_MOUSE_CALLBACK) )
		{
			notice("Failed to hook");
			goto E_BEGIN;
		}

		return TRUE;
	E_BEGIN:
	E_PROC:
		FreeLibrary(G_HookDLLModule);
		G_HookDLLModule = NULL;
	E_LOAD:
	E_PATH:
		return FALSE;
	}
}


static BOOL
engine_end(void)
{
	if (G_Engine == TIMER) {
		return KillTimer(G_PluginWnd, IDT_CHECK) != 0;
	} else {
		FARPROC end;
		BOOL result;

		end = GetProcAddress(G_HookDLLModule, "BBHSMouseHook_End");
		assert(end != NULL);

		result = (*((BOOL (*)(void))end))();

		FreeLibrary(G_HookDLLModule);
		G_HookDLLModule = NULL;

		if (!result)
			notice("Failed to unhook");
		return result;
	}
}




static void
interpret_broam(const char* broam)
{
	const char* token[8];
	int token_length[NUMBER_OF(token)];
	int token_num;
	const char* rest;
	char tmp[80];
	HotSpot* hotspot;

	#define IS(s) ( (NUMBER_OF(s) - 1 == token_length[1]) \
	                && !strncmp((s), token[1], token_length[1]) )

	/* 0:@BBHotSpots, 1:CMD, rest:ARGS */
	tokenize(broam, token, token_length, &token_num, 2, &rest);
	if (token_num != 2)
		return;
	if ( (token_length[0] != 11)
	     || (strncmp("@BBHotSpots", token[0], token_length[0])) )
	{
		return;
	}

	if (IS("load_settings")) {
		/**
		 * Calling read_settings() in here
		 * causes the core to crash or be deadlock.
		 * So this bro@m should be POST a message
		 * that requests to call read_settings().
		 */
		PostMessage(G_PluginWnd, BB_RECONFIGURE, 0, 0);
	} else if (IS("engine")) {
		int new_engine;

		if (!stricmp(rest, "timer"))
			new_engine = TIMER;
		else if (!stricmp(rest, "hook"))
			new_engine = HOOK;
		else {
			notice( "%s\n\nEngine must be `timer' or `hook'.",
			        broam );
			return;
		}

		if (new_engine == G_Engine)
			return;

		if (new_engine == TIMER) {
			engine_end();
			G_Engine = TIMER;
			engine_begin();
		} else {
			engine_end();
			G_Engine = HOOK;
			engine_begin();
		}
	} else if (IS("interval")) {
		if (G_Engine == TIMER) {
		  G_CheckingInterval = atoi(rest);
		  if (G_CheckingInterval < 100)
		    G_CheckingInterval = 100;
		  SetTimer(G_PluginWnd, IDT_CHECK, G_CheckingInterval, NULL);
		}
	} else if (IS("clear")) {
		HotSpotList_Del(G_HotSpotList);
		G_HotSpotList = HotSpotList_New();
	} else if (IS("add") || IS("set")) {
		BOOL is_add = (token[1][0] == 'a');
		RECT spot;
		BOOL active;

		/* 0:name, 1:x, 2:y, 3:width, 4:height, (5:state) */
		tokenize(rest, token, token_length, &token_num, 6, &rest);
		if ((token_num < 5) || (6 < token_num)) {
			notice("%s\n\nSyntax error.", broam);
			return;
		}
		if (token_length[0] == 0) {
			notice("%s\n\n`name' cannot be null-string.", broam);
			return;
		}

		spot.right = convert_width(token[3], token_length[3]);
		spot.bottom = convert_height(token[4], token_length[4]);
		spot.left = convert_x(token[1], token_length[1], spot.right);
		spot.top = convert_y(token[2], token_length[2], spot.bottom);
		spot.right += spot.left;
		spot.bottom += spot.top;
		active = ( token_num == 6
		           ? convert_bool(token[5], token_length[5])
		           : TRUE );

		if (is_add) {
			int result;

			result = HotSpotList_Add( &G_HotSpotList,
			                          token[0], token_length[0],
			                          &spot, active );
			if (result == NO_ERROR)
				return;

			STRNCPY(tmp, token[0], token_length[0]+1);
			if (result == ERROR_OUTOFMEMORY) {
			  notice("Out of memory to add hotspot `%s'", tmp);
			} else if (result == ERROR_ALREADY_EXISTS) {
			  notice("Hotspot `%s' is already defined", tmp);
			} else {
			  notice("Unknown error occured at adding `%s'", tmp);
			}
		} else {
			Token param;
			int n;

			param.string = token[0];
			param.length = token_length[0];
			n = HotSpotList_ForEach( G_HotSpotList,
			                         HotSpot_MatchByName, &param,
			                         HotSpot_SetSpot, &spot );
			if (n == 0) {
				STRNCPY(tmp, token[0], token_length[0]+1);
				notice("No hotspot named `%s'", tmp);
				return;
			}
		}
	} else if (IS("del")) {
		tokenize(rest, token, token_length, &token_num, 1, &rest);
		if ( !HotSpotList_DelByName
		       (&G_HotSpotList, token[0], token_length[0]) )
		{
			notice("No hotspot named `%s'", token[0]);
		}
	} else if (IS("on-enter") || IS("on-leave") || IS("on-click")) {
		char type = token[1][3];
		int line;
		int column;
		const char* message;
		BBCmdCommand* cmd;

		/* 0:name, rest:script */
		tokenize(rest, token, token_length, &token_num, 1, &rest);
		if (token_num != 1) {
			notice("%s\n\nSyntax error.", broam);
			return;
		}

		hotspot = HotSpotList_SearchByName(
		            G_HotSpotList,
		            token[0], token_length[0]
		          );
		if (hotspot == NULL) {
			STRNCPY(tmp, token[0], token_length[0]+1);
			notice("No hotspot named `%s'", tmp);
			return;
		}

		cmd = compile(rest, &line, &column, &message);
		if (cmd == NULL) {
			notice( "%s\n"
			        "\n"
			        "Compile error.\n"
			        "Line %d, column %d: %s",
			        broam,
			        line, column, message );
			return;
		}

		if (type == 'e') {  /* on-enter */
			BBCmdCommand_Del(hotspot->enter_script);
			hotspot->enter_script = cmd;
		} else if (type == 'l') {  /* on-leave */
			BBCmdCommand_Del(hotspot->leave_script);
			hotspot->leave_script = cmd;
		} else {  /* if (type == 'c')  # on-click */
			BBCmdCommand_Del(hotspot->click_script);
			hotspot->click_script = cmd;
		}
	} else if ( IS("activate") || IS("inactivate") || IS("toggle")
	            || IS("info") )
	{
		HotSpotList_ForEachProc proc;
		Token param;
		int n;

		if (token[1][2] == 't')
			proc = HotSpot_Activate;
		else if (token[1][2] == 'a')
			proc = HotSpot_Inactivate;
		else if (token[1][2] == 'g')
			proc = HotSpot_Toggle;
		else /* if (token[1][2] == 'f') */
			proc = HotSpot_Info;

		tokenize(rest, token, token_length, &token_num, 1, &rest);
		param.string = token[0];
		param.length = (token_num == 0 ? 0 : token_length[0]);
		n = HotSpotList_ForEach( G_HotSpotList,
		                         HotSpot_MatchByName, &param,
		                         proc, (void*)(param.length == 0) );
		if ((n == 0) && (token_num != 0)) {
			STRNCPY(tmp, token[0], token_length[0]+1);
			notice("No hotspot named `%s'", tmp);
		}
	} else {
		notice("%s\n\nUnknown bro@m.", broam);
	}

	#undef IS
}




static void
check(UINT msg, POINT pt)
{
	HotSpot* hotspot;

	if ((msg == WM_LBUTTONUP)
	    || (msg == WM_RBUTTONUP)
	    || (msg == WM_MBUTTONUP)
	    || (msg == WM_XBUTTONUP)
	    || (msg == WM_NCLBUTTONUP)
	    || (msg == WM_NCRBUTTONUP)
	    || (msg == WM_NCMBUTTONUP)
	    || (msg == WM_NCXBUTTONUP)) {  /* is click message? */
	  if (G_LastActiveHotSpot != NULL) {
	    if (PtInRect(&(G_LastActiveHotSpot->spot), pt)) {
	        if (G_LastActiveHotSpot->active
	            && (G_LastActiveHotSpot->click_script != NULL))
	        {
	          BBCmdCommand_Execute(G_LastActiveHotSpot->click_script);
	        }
	    }
	  }
	  return;
	}

	if ((G_LastPosition.x == pt.x) && (G_LastPosition.y == pt.y))
		return;
	G_LastPosition = pt;

	if (G_LastActiveHotSpot != NULL) {
	  if (!PtInRect(&(G_LastActiveHotSpot->spot), pt)) {
	    if ( G_LastActiveHotSpot->active
	         && (G_LastActiveHotSpot->leave_script != NULL) )
	    {
	      BBCmdCommand_Execute(G_LastActiveHotSpot->leave_script);
	    }
	    G_LastActiveHotSpot = NULL;
	  }
	}

	if (G_LastActiveHotSpot == NULL) {
	  hotspot = G_HotSpotList;
	  while (hotspot != NULL) {
	    if (PtInRect(&(hotspot->spot), pt)) {
	      if (hotspot->active && (hotspot->enter_script != NULL))
	        BBCmdCommand_Execute(hotspot->enter_script);
	      G_LastActiveHotSpot = hotspot;
	      break;
	    }

	    hotspot = hotspot->next;
	  }
	}
}




static LRESULT CALLBACK
BBHS_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case BB_BROADCAST:
		interpret_broam((const char*)lp);
		break;

	case BB_RECONFIGURE:
		update_screen_rect();
		read_settings();
		break;


	case WH_MOUSE_CALLBACK: {
		POINT pt;

		pt.x = LOWORD(lp);
		pt.y = HIWORD(lp);
		check(wp, pt);
		} break;

	case WM_TIMER:
		if (wp == IDT_CHECK) {
			POINT pt;

			if (GetCursorPos(&pt))
				check(WM_MOUSEMOVE, pt);
		}
		break;


	case WM_CREATE:
		SendMessage( GetBBWnd(), BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBHS_MESSAGES );
		break;

	case WM_DESTROY:
		SendMessage( GetBBWnd(), BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBHS_MESSAGES );
		break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}








/* Public functions */

DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	G_PluginModule = (HMODULE)hplugininstance;
	update_screen_rect();

	if ( BBCmd_IsCompatible(BBCmd_VERSION, BBCmd_ICH_JUST_USE)
	     != BBCmd_IC_COMPATIBLE )
	{
		notice("Imcompatible BBCmd");
		goto E_BBCMD_COMPATIBILITY;
	}

	G_HotSpotList = HotSpotList_New();

	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = BBHS_WndProc;
	wc.hInstance = hplugininstance;
	wc.lpszClassName = BBHS_CLASS_NAME;
	if (!RegisterClass(&wc)) {
		notice("Failed RegisterClass()");
		goto E_REGISTER_CLASS;
	}

	G_PluginWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                              BBHS_CLASS_NAME,
	                              NULL,
	                              0,
	                              0, 0, 0, 0,
	                              (is_2k_or_later() ? HWND_MESSAGE : NULL),
	                              NULL,
	                              hplugininstance,
	                              NULL );
	if (G_PluginWnd == NULL) {
		notice("Failed CreateWindowEx()");
		goto E_CREATE_WINDOW;
	}

	if (!engine_begin())
		goto E_ENGINE_BEGIN;

	if (!read_settings())
		goto E_READ_SETTINGS;

	return 0;


E_READ_SETTINGS:
	engine_end();
E_ENGINE_BEGIN:
	DestroyWindow(G_PluginWnd);
E_CREATE_WINDOW:
	UnregisterClass(BBHS_CLASS_NAME, hplugininstance);
E_REGISTER_CLASS:
	HotSpotList_Del(G_HotSpotList);
E_BBCMD_COMPATIBILITY:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	engine_end();

	DestroyWindow(G_PluginWnd);
	UnregisterClass(BBHS_CLASS_NAME, hplugininstance);

	HotSpotList_Del(G_HotSpotList);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	default: /* FALLTHRU */
	case PLUGIN_NAME: return BBHS_NAME;
	case PLUGIN_VERSION: return BBHS_VERSION;
	case PLUGIN_AUTHOR: return BBHS_AUTHOR;
	case PLUGIN_RELEASE: return BBHS_RELEASE;
	case PLUGIN_LINK: return BBHS_LINK;
	case PLUGIN_EMAIL: return BBHS_EMAIL;
	}
}




/* __END__ */
